// use crate::process::{LErr, Trace};
use crate::parser::Ast;
use crate::spans::Spanned;
use crate::validator::VErr;
use crate::value::Value;
use crate::vm::CallFrame;
use ariadne::{sources, Color, Label, Report, ReportKind};
use std::collections::HashSet;

#[derive(Debug, Clone, PartialEq)]
pub enum PanicInfo {
    RepeatNumberType(Value),
    NoMatch,
    LetMatchFail,
    MatchFail,
    FnMatchFail,
    Internal(&'static str),
    ListSplat(Value),
    DictSplat(Value),
    DecType(Value),
    WhenFail,
    AddType(Value, Value),
    SubType(Value, Value),
    MultType(Value, Value),
    DivType(Value, Value),
    UnboxType(Value),
    StoreType(Value),
    Assert,
    Explicit(Value),
    GetType(Value),
    AtIdxType(Value),
    AtCollectionType(Value),
    PartialType(Value),
    ArityMismatch(Value, Vec<Value>),
    CallType(Value),
    DivisionByZero,
}

impl PanicInfo {
    pub fn msg(&self) -> String {
        use PanicInfo::*;
        match self {
            RepeatNumberType(value) => format!("`repeat` expects a number; it got `{value}`"),
            _ => todo!(),
        }
    }
}

#[derive(Debug, Clone, PartialEq)]
pub struct Panic {
    pub frame: CallFrame,
    pub call_stack: Vec<CallFrame>,
    pub info: PanicInfo,
}

impl Panic {
    pub fn report(&self) {
        println!("got frame: {}", self.frame);
        let mut srcs = HashSet::new();
        let ast_spans = std::iter::zip(
            self.frame.chunk().spans.iter(),
            self.frame.chunk().nodes.iter(),
        );
        let ast_spans = ast_spans
            .map(|(span, ast)| format!("{span} || {}", ast.show()))
            .collect::<Vec<_>>()
            .join("\n");
        println!("{ast_spans}");
        let msg = self.info.msg();
        let input = self.frame.chunk().input;
        let src = self.frame.chunk().src;
        srcs.insert((input, src));
        let mut traceback = format!("Ludus panicked! {msg}");
        for frame in self.call_stack.iter().rev() {
            let name = frame.function.as_fn().name();
            let args = frame
                .args
                .iter()
                .map(|arg| arg.show())
                .collect::<Vec<_>>()
                .join(", ");
            traceback = format!("{traceback}\n  calling `{name}` with ({args})");
        }
        println!("{traceback}");
    }
}

// pub fn report_panic(err: LErr) {
//     let mut srcs = HashSet::new();
//     let mut stack = vec![];
//     let mut order = 1;
//     for entry in err.trace.iter().rev() {
//         let Trace {
//             callee,
//             caller,
//             function,
//             arguments,
//             input,
//             src,
//         } = entry;
//         let (_, first_span) = callee;
//         let (_, second_span) = caller;
//         let Value::Fn(f) = function else {
//             unreachable!()
//         };
//         let fn_name = f.borrow().name.clone();
//         let i = first_span.start;
//         let j = second_span.end;
//         let label = Label::new((entry.input, i..j))
//             .with_color(Color::Yellow)
//             .with_message(format!("({order}) calling `{fn_name}` with `{arguments}`"));
//         order += 1;
//         stack.push(label);
//         srcs.insert((*input, *src));
//     }

//     Report::build(ReportKind::Error, (err.input, err.span.into_range()))
//         .with_message(format!("Ludus panicked! {}", err.msg))
//         .with_label(Label::new((err.input, err.span.into_range())).with_color(Color::Red))
//         .with_labels(stack)
//         .with_note(err.extra)
//         .finish()
//         .print(sources(srcs.iter().copied()))
//         .unwrap();
// }

pub fn report_invalidation(errs: Vec<VErr>) {
    for err in errs {
        Report::build(ReportKind::Error, (err.input, err.span.into_range()))
            .with_message(err.msg.to_string())
            .with_label(Label::new((err.input, err.span.into_range())).with_color(Color::Cyan))
            .finish()
            .print(sources(vec![(err.input, err.src)]))
            .unwrap();
    }
}
